*************************************************************************************
*														*
*	BATTLEZONE type volcano for the EASy68k simulator	2005/06/12			*
*														*
*	This code uses double buffering to render a BATTLEZONE type volcano		*
*	effect.The effect is not an end to itself but just an idea of how such		*
*	an effect can be implemented. The random number generator used is very		*
*	similar to that used in EhBASIC68.								*
*														*
*	This version for Sim68K 3.2 or later							*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	 http://www.themotionstore.com/leeedavison/index.html					*
*														*
*	.. or ..												*
*														*
*	 http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************

	ORG	$1000

start
	LEA		(LAB_VARS,PC),a3		* get pointer to variables
	MOVE.l	#$8C9F53D0,PRNlword(a3)	* set PRNG seed. this can be any value
							* except zero, changing this will produce
							* a different start point for the effect

	BSR		Clear_stars			* clear the volcano stars

	MOVE.b	#17,d1			* enable double buffering
	MOVE.b	#92,d0			* set draw mode
	TRAP		#15

	MOVEQ		#11,d0			* position cursor
	MOVE.w	#$FF00,d1			* clear screen
	TRAP		#15

loop
	MOVEQ		#94,d0			* copy buffer to main
	TRAP		#15

	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	MOVE.l	d1,d7				* copy time to d7

	MOVEQ		#11,d0			* position cursor
	MOVE.w	#$FF00,d1			* clear screen
	TRAP		#15

	BSR		Update_stars		* go update the star data
	BSR		Draw_background		* draw`the background
	BSR		Draw_stars			* draw the volcano stars

* now wait for a while. there is enough time for many more stars as can be seen by
* running without the delay, but is set like this to simulate the BATTLEZONE type
* volcano more closely

wait_100ms
	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	SUB.l		d7,d1				* subtract previous time from current time
	CMP.b		#$09,d1			* compare with 9/100ths
	BMI.s		wait_100ms			* loop if time not up yet

	BRA		loop				* loop forever

*************************************************************************************
* This is the code that generates the pseudo random sequence. A seed word located in
* PRNlword(a3) is loaded into a register before being operated on to generate the
* next number in the sequence. This number is then saved as the seed for the next
* time it's called.
*
* This code is adapted from the 32 bit version of RND(n) used in EhBASIC68. Taking
* the 19th next number is slower but helps to hide the shift and add nature of this
* generator as can be seen from analysing the output.

LAB_RND
	MOVEM.l	d0-d2,-(sp)			* save d0, d1 and d2
	MOVE.l	PRNlword(a3),d0		* get current seed longword
	MOVEQ		#$AF-$100,d1		* set EOR value
	MOVEQ		#18,d2			* do this 19 times
Ninc0
	ADD.l		d0,d0				* shift left 1 bit
	BCC.s		Ninc1				* if bit not set skip feedback

	EOR.b		d1,d0				* do Galois LFSR feedback
Ninc1
	DBF		d2,Ninc0			* loop

	MOVE.l	d0,PRNlword(a3)		* save back to seed longword
	MOVEM.l	(sp)+,d0-d2			* restore d0, d1 and d2

	RTS

*************************************************************************************
*
* star update routine. this routine spawns new stars if there are extinguished stars
* and updates the position and velocity of the current stars. much use is made of the
* random number generator to spawn new stars.

Update_stars
	MOVEQ		#16,d6			* set star count * 4
Update_loop
	MOVE.l	st_intens(a3,d6),d0	* get star intensity
	BNE		St_visible			* branch if star visible

	BSR		LAB_RND			* call the PRNG code
	MOVE.b	PRNlword(a3),d0		* get byte from result
	AND.b		#$07,d0			* mask 1 of 8
	BNE		St_do_next			* no new star 7/8 ths of the time

	MOVE.l	#$0000F8,d1			* red pen
	MOVE.l	d1,st_intens(a3,d6)	* save star intensity

	BSR		LAB_RND			* call the PRNG code
	MOVE.b	PRNlword(a3),d1		* get byte from result
	AND.w		#3,d1				* mask 0 to 3
	ADDQ.w	#1,d1				* make 1 - 4
	BSR		LAB_RND			* call the PRNG code
	MOVE.b	PRNlword(a3),d0		* get byte from result
	AND.b		#$40,d0			* mask bit 6
	BEQ		St_goright			* branch if bit was zero

	NEG.w		d1				* else make delta x left
St_goright
	MOVE.w	d1,st_deltax(a3,d6)	* save star delta x

	BSR		LAB_RND			* call the PRNG code
	MOVE.b	PRNlword(a3),d1		* get byte from result
	AND.w		#7,d1				* mask 0 to 7
	ADDQ.w	#5,d1				* make 5 - 12
	MOVE.w	d1,st_deltay(a3,d6)	* save star delta y
	MOVEQ		#$5D,d0			* set start y co-ordinate
St_gone
	MOVE.w	d0,st_ycoord(a3,d6)	* save star y co-ordinate
	MOVEQ		#0,d0				* clear register
	MOVE.w	d0,st_xcoord(a3,d6)	* clear star x co-ordinate
	BRA		St_do_next			* go do next star

St_visible
	SUBQ.l	#8,d0				* decrement star intensity
	MOVE.l	d0,st_intens(a3,d6)	* save star intensity
	BEQ		St_gone			* if star extinguished go clear (x,y)

	MOVE.w	st_deltay(a3,d6),d0	* get star delta y
	SUBQ.w	#1,d0				* decrement star delta y
	MOVE.w	d0,st_deltay(a3,d6)	* save new star delta y
	ADD.w		st_ycoord(a3,d6),d0	* add star y co-ordinate
	BPL		St_insky			* branch if star is above horizon

	MOVEQ		#0,d0				* clear register
	MOVE.l	d0,st_intens(a3,d6)	* turn star off
	BRA		St_gone			* go clear (x,y)

* star is above the horizon

St_insky
	MOVE.w	d0,st_ycoord(a3,d6)	* save star y co-ordinate

	MOVE.w	st_deltax(a3,d6),d0	* get star delta x
	ADD.w		st_xcoord(a3,d6),d0	* add star x co-ordinate
	MOVE.w	d0,st_xcoord(a3,d6)	* save star x co-ordinate

St_do_next
	SUBQ.w	#4,d6				* decrement index
	BPL		Update_loop			* loop if not all done

	RTS

*************************************************************************************
*
* star draw routine. checks each star in turn and draws it if it's visible

Draw_stars
	MOVEQ		#16,d6			* set star count * 4
Draw_loop
	MOVE.l	st_intens(a3,d6),d1	* get star intensity
	BEQ		St_invisible		* branch if star not visible

	MOVEQ		#80,d0			* set pen colour to star colour
	TRAP		#15

	MOVE.w	#400,d1			* set x origin to screen right(ish)
	ADD.w		st_xcoord(a3,d6),d1	* add star x co-ordinate

	MOVE.w	#240,d2			* set y origin to screen centre
	SUB.w		st_ycoord(a3,d6),d2	* add star y co-ordinate, note y = 0 at top

	MOVE.w	d1,d3				* copy x co-ordinate
	ADDQ.w	#1,d3				* make a small square

	MOVE.w	d2,d4				* copy y co-ordinate
	ADDQ.w	#1,d4				* make a small square


	MOVEQ		#87,d0			* draw rectangle
	TRAP		#15
St_invisible
	SUBQ.w	#4,d6				* decrement index
	BPL		Draw_loop			* loop if not all done

	RTS

*************************************************************************************
*
* star clear routine. initialise the stars as memory isn't always $00000000 when you
* expect it to be

Clear_stars
	MOVEQ		#16,d6			* set star count * 4
	MOVEQ		#0,d0				* clear register
Clear_loop
	MOVE.l	d0,st_intens(a3,d6)	* clear the star intensity
	SUBQ.w	#4,d6				* decrement index
	BPL		Clear_loop			* loop if not all done

	RTS

*************************************************************************************
*
* draw the, somewhat boring, background

Draw_background
	MOVE.l	#$00F800,d1			* green pen
	MOVEQ		#80,d0			* set pen colour
	TRAP		#15

	LEA		background(PC),a4	* get pointer to background data
Bg_loop
	MOVE.w	(a4)+,d1			* get x origin
	BMI.s		Bg_done			* exit if done

	MOVE.w	(a4)+,d2			* get y origin
	MOVE.w	(a4)+,d3			* get x end
	MOVE.w	(a4)+,d4			* get y end

	MOVEQ		#84,d0			* draw line
	TRAP		#15

	BRA.s		Bg_loop			* go do next line

Bg_done
	RTS

background
	dc.w		  0,240,640,240		* horizon
	dc.w		360,240,390,147		* left slope
	dc.w		390,147,395,157		* dip
	dc.w		395,157,400,150		* peak
	dc.w		400,150,405,157		* dip
	dc.w		405,157,410,147		* peak
	dc.w		410,147,440,240		* right slope
	dc.w		-1				* end marker


*************************************************************************************
*
* variables used

LAB_VARS

	OFFSET	0				* going to use relative addressing

PRNlword	ds.l	1				* PRNG seed long word

* there are five volcano stars indexed as 0 to 4. intensity also sets the colour. delta y
* is 5 to 12 and decremented each loop to simulate acceleration due to gravity. delta x is
* -4 to +4 but not 0. -ve values are to the left. (x,y) are the 2D co-ordinate pair, the
* horizon is the lower y limit

st_intens	ds.l	5				* volcano star intensity
st_deltax	ds.l	5				* volcano star delta x
st_deltay	ds.l	5				* volcano star delta y
st_xcoord	ds.l	5				* volcano star x co-ordinate
st_ycoord	ds.l	5				* volcano star y co-ordinate

	ORG	*

	END	start

